import { __extends } from "tslib";
import { Type } from './Type';
import { cloneObject, createCatchForKey, createValuesExtractor, isArray, isEqual, isEqualArray, isFast, promiseAll, promiseAllSettled, raiseIssue, raiseIssuesIfDefined, raiseIssuesOrCaptureForKey, raiseIssuesOrPush, } from '../utils';
/**
 * The array type definition.
 *
 * @template X The type definition of array elements.
 */
var ArrayType = /** @class */ (function (_super) {
    __extends(ArrayType, _super);
    /**
     * Creates the new {@link ArrayType} instance.
     *
     * @param type The type definition of array elements.
     * @param options The type constraint options.
     */
    function ArrayType(type, options) {
        var _this = _super.call(this, type.async, options) || this;
        _this.type = type;
        return _this;
    }
    /**
     * Constrains the array length.
     *
     * @param length The length of the array to satisfy the constraint.
     * @param options The options applied to both {@link min} and {@link max} constraints.
     * @returns The new type definition.
     */
    ArrayType.prototype.length = function (length, options) {
        return this.min(length, options).max(length, options);
    };
    /**
     * Constrains the minimum array length.
     *
     * @param length The minimum length of the array to satisfy the constraint.
     * @param options The minimum constraint options.
     * @returns The new type definition.
     */
    ArrayType.prototype.min = function (length, options) {
        var type = cloneObject(this);
        type.minLength = length;
        type.minLengthOptions = options;
        return type;
    };
    /**
     * Constrains the maximum array length.
     *
     * @param length The maximum length of the array to satisfy the constraint.
     * @param options The maximum constraint options.
     * @returns The new type definition.
     */
    ArrayType.prototype.max = function (length, options) {
        var type = cloneObject(this);
        type.maxLength = length;
        type.maxLengthOptions = options;
        return type;
    };
    ArrayType.prototype.parse = function (input, options) {
        if (!isArray(input)) {
            raiseIssue(input, 'type', 'array', this.options, 'Must be an array');
        }
        var _a = this, minLength = _a.minLength, maxLength = _a.maxLength, type = _a.type;
        var inputLength = input.length;
        var issues;
        if (minLength != null && inputLength < minLength) {
            issues = raiseIssuesOrPush(issues, options, input, 'arrayMinLength', minLength, this.minLengthOptions, 'Must have the minimum length of ' + minLength);
        }
        if (maxLength != null && inputLength > maxLength) {
            issues = raiseIssuesOrPush(issues, options, input, 'arrayMaxLength', maxLength, this.maxLengthOptions, 'Must have the maximum length of ' + maxLength);
        }
        if (this.async) {
            var promises = [];
            var handleOutput = function (output) { return (isEqualArray(input, output) ? input : output); };
            for (var i = 0; i < inputLength; ++i) {
                promises.push(type.parse(input[i], options).catch(createCatchForKey(i)));
            }
            if (isFast(options)) {
                return promiseAll(promises).then(handleOutput);
            }
            return promiseAllSettled(promises).then(createValuesExtractor(issues)).then(handleOutput);
        }
        var output = input;
        for (var i = 0; i < inputLength; ++i) {
            var value = input[i];
            var outputValue = void 0;
            try {
                outputValue = type.parse(value, options);
            }
            catch (error) {
                issues = raiseIssuesOrCaptureForKey(error, issues, options, i);
            }
            if (isEqual(outputValue, value) || issues !== undefined) {
                continue;
            }
            if (output === input) {
                output = input.slice(0);
            }
            output[i] = outputValue;
        }
        raiseIssuesIfDefined(issues);
        return output;
    };
    return ArrayType;
}(Type));
export { ArrayType };
