import { __extends } from "tslib";
import { Type } from './Type';
import { cloneObject, raiseIssue, raiseIssuesIfDefined, raiseIssuesOrPush } from '../utils';
var isFinite = Number.isFinite;
/**
 * The number type definition.
 */
var NumberType = /** @class */ (function (_super) {
    __extends(NumberType, _super);
    function NumberType(options) {
        return _super.call(this, false, options) || this;
    }
    /**
     * Constrains the number to be greater than zero.
     */
    NumberType.prototype.positive = function (options) {
        return this.gt(0, options);
    };
    /**
     * Constrains the number to be less than zero.
     */
    NumberType.prototype.negative = function (options) {
        return this.lt(0, options);
    };
    /**
     * Constrains the number to be greater than the value.
     */
    NumberType.prototype.gt = function (value, options) {
        var type = cloneObject(this);
        type.min = value;
        type.minIncluded = false;
        type.minOptions = options;
        return type;
    };
    /**
     * Constrains the number to be less than the value.
     */
    NumberType.prototype.lt = function (value, options) {
        var type = cloneObject(this);
        type.max = value;
        type.maxIncluded = false;
        type.maxOptions = options;
        return type;
    };
    /**
     * Constrains the number to be greater than or equal to the value.
     */
    NumberType.prototype.gte = function (value, options) {
        var type = cloneObject(this);
        type.min = value;
        type.minIncluded = true;
        type.minOptions = options;
        return type;
    };
    /**
     * Constrains the number to be less than or equal to the value.
     */
    NumberType.prototype.lte = function (value, options) {
        var type = cloneObject(this);
        type.max = value;
        type.maxIncluded = true;
        type.maxOptions = options;
        return type;
    };
    /**
     * Constrains the number to be a multiple of the divisor.
     *
     * @param divisor The number by which the input should be divisible without a remainder.
     * @param options
     */
    NumberType.prototype.multipleOf = function (divisor, options) {
        var type = cloneObject(this);
        type.divisor = divisor;
        type.divisorOptions = options;
        return type;
    };
    NumberType.prototype.parse = function (input, options) {
        if (!isFinite(input)) {
            raiseIssue(input, 'type', 'number', this.options, 'Must be a number');
        }
        var _a = this, min = _a.min, max = _a.max, minIncluded = _a.minIncluded, maxIncluded = _a.maxIncluded, divisor = _a.divisor;
        var issues;
        if (min != null && (minIncluded ? input < min : input <= min)) {
            issues = raiseIssuesOrPush(issues, options, input, minIncluded ? 'numberGreaterThanOrEqual' : 'numberGreaterThan', min, this.minOptions, 'Must be greater than ' + (minIncluded ? 'or equal to ' + min : min));
        }
        if (max != null && (maxIncluded ? input > max : input >= max)) {
            issues = raiseIssuesOrPush(issues, options, input, maxIncluded ? 'numberLessThanOrEqual' : 'numberLessThan', max, this.maxOptions, 'Must be less than ' + (maxIncluded ? 'or equal to ' + max : max));
        }
        if (divisor != null && input % divisor !== 0) {
            issues = raiseIssuesOrPush(issues, options, input, 'numberMultipleOf', divisor, this.divisorOptions, 'Must be a multiple of ' + divisor);
        }
        raiseIssuesIfDefined(issues);
        return input;
    };
    return NumberType;
}(Type));
export { NumberType };
