import { AnyType, InferType, Type } from './Type';
import { Awaitable, ConstraintOptions, Dict, ParserOptions, Several } from '../shared-types';
declare type InferObjectType<P extends Dict<AnyType>, I extends AnyType> = Squash<UndefinedAsOptional<{
    [K in keyof P]: InferType<P[K]>;
}> & InferIndexerType<I>>;
declare type InferIndexerType<I extends AnyType> = I extends Type<never> ? unknown : {
    [indexer: string]: InferType<I>;
};
declare type Squash<T> = T extends never ? never : {
    [K in keyof T]: T[K];
};
declare type UndefinedAsOptional<T> = OmitBy<T, undefined> & Partial<PickBy<T, undefined>>;
declare type OmitBy<T, V> = Omit<T, {
    [K in keyof T]: V extends Extract<T[K], V> ? K : never;
}[keyof T]>;
declare type PickBy<T, V> = Pick<T, {
    [K in keyof T]: V extends Extract<T[K], V> ? K : never;
}[keyof T]>;
declare const enum ObjectKeysMode {
    EXACT = 0,
    STRIP = 1,
    PRESERVE = 2,
    INDEXER = 3
}
/**
 * The object type definition.
 *
 * @template P The mapping from an object key to a corresponding type definition.
 * @template I The type definition that constrains the indexer signature.
 */
export declare class ObjectType<P extends Dict<AnyType>, I extends AnyType> extends Type<InferObjectType<P, I>> {
    protected props: P;
    protected indexerType: I | null;
    protected keys: string[];
    protected valueTypes: AnyType[];
    protected keysMode: ObjectKeysMode;
    protected propEntries: [string, AnyType][];
    protected exactOptions?: ConstraintOptions;
    /**
     * Creates a new {@link ObjectType} instance.
     *
     * @param props The mapping from an object key to a corresponding type definition.
     * @param indexerType The type definition that constrains the indexer signature. If `null` then values thea fall
     * under the indexer signature are unconstrained.
     * @param options
     */
    constructor(props: P, indexerType: I | null, options?: ConstraintOptions);
    /**
     * Merge properties from the other object type. If a property with the same key already exists on this object type
     * then it is overwritten. Indexer signature of this type is preserved intact.
     *
     * @param type The object type which properties must be added to this object type.
     * @returns The modified object type.
     *
     * @template P1 The type of properties to add.
     */
    extend<P1 extends Dict<AnyType>>(type: ObjectType<P1, AnyType>): ObjectType<Pick<P, Exclude<keyof P, keyof P1>> & P1, I>;
    /**
     * Add properties to an object type. If a property with the same key already exists on this object type then it is
     * overwritten.
     *
     * @param props The properties to add.
     * @returns The modified object type.
     *
     * @template P1 The type of properties to add.
     */
    extend<P1 extends Dict<AnyType>>(props: P1): ObjectType<Pick<P, Exclude<keyof P, keyof P1>> & P1, I>;
    /**
     * Returns an object type that only has properties with listed keys.
     *
     * @param keys The list of property keys to pick.
     * @returns The modified object type.
     *
     * @template K The tuple of keys to pick.
     */
    pick<K extends Several<keyof P & string>>(...keys: K): ObjectType<Pick<P, K[number]>, I>;
    /**
     * Returns an object type that doesn't have the listed keys.
     *
     * @param keys2 The list of property keys to omit.
     * @returns The modified object type.
     *
     * @template K The tuple of keys to omit.
     */
    omit<K extends Several<keyof P & string>>(...keys2: K): ObjectType<Omit<P, K[number]>, I>;
    /**
     * Returns an object type that allows only known keys and has no index signature.
     *
     * @returns The modified object type.
     */
    exact(options?: ConstraintOptions): ObjectType<P, Type<never>>;
    /**
     * Returns an object type that doesn't have indexer signature and all unknown keys are stripped.
     *
     * @returns The modified object type.
     */
    strip(): ObjectType<P, Type<never>>;
    /**
     * Returns an object type that has an indexer signature that doesn't constrain values.
     *
     * @returns The modified object type.
     */
    preserve(): ObjectType<P, Type<any>>;
    /**
     * Returns an object type that has an indexer signature that is constrained by the given type.
     *
     * @param indexType The type of the indexer values.
     * @returns The modified object type.
     *
     * @template I1 The indexer signature type.
     */
    index<I1 extends AnyType>(indexType: I1): ObjectType<P, I1>;
    parse(input: unknown, options?: ParserOptions): Awaitable<InferObjectType<P, I>>;
}
export {};
