import { __extends } from "tslib";
import { Type } from './Type';
import { cloneObject, copyObjectEnumerableKeys, copyObjectKnownKeys, createCatchForKey, createValuesExtractor, isAsync, isEqual, isFast, isObjectLike, parseAsync, promiseAll, promiseAllSettled, raiseIssue, raiseIssuesIfDefined, raiseIssuesOrCaptureForKey, raiseIssuesOrPush, } from '../utils';
/**
 * The object type definition.
 *
 * @template P The mapping from an object key to a corresponding type definition.
 * @template I The type definition that constrains the indexer signature.
 */
var ObjectType = /** @class */ (function (_super) {
    __extends(ObjectType, _super);
    /**
     * Creates a new {@link ObjectType} instance.
     *
     * @param props The mapping from an object key to a corresponding type definition.
     * @param indexerType The type definition that constrains the indexer signature. If `null` then values thea fall
     * under the indexer signature are unconstrained.
     * @param options
     */
    function ObjectType(props, indexerType, options) {
        var _this = this;
        var valueTypes = Object.values(props);
        _this = _super.call(this, (indexerType === null || indexerType === void 0 ? void 0 : indexerType.async) || isAsync(valueTypes), options) || this;
        _this.props = props;
        _this.indexerType = indexerType;
        _this.keys = Object.keys(props);
        _this.valueTypes = valueTypes;
        _this.propEntries = Object.entries(props);
        _this.keysMode = indexerType === null ? 2 /* ObjectKeysMode.PRESERVE */ : 3 /* ObjectKeysMode.INDEXER */;
        return _this;
    }
    ObjectType.prototype.extend = function (arg) {
        var nextProps = Object.assign({}, this.props, arg instanceof ObjectType ? arg.props : arg);
        var type = new ObjectType(nextProps, this.indexerType);
        type.keysMode = this.keysMode;
        return type;
    };
    /**
     * Returns an object type that only has properties with listed keys.
     *
     * @param keys The list of property keys to pick.
     * @returns The modified object type.
     *
     * @template K The tuple of keys to pick.
     */
    ObjectType.prototype.pick = function () {
        var keys = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            keys[_i] = arguments[_i];
        }
        var nextProps = {};
        for (var _a = 0, keys_1 = keys; _a < keys_1.length; _a++) {
            var key = keys_1[_a];
            nextProps[key] = this.props[key];
        }
        var type = new ObjectType(nextProps, this.indexerType);
        type.keysMode = this.keysMode;
        return type;
    };
    /**
     * Returns an object type that doesn't have the listed keys.
     *
     * @param keys2 The list of property keys to omit.
     * @returns The modified object type.
     *
     * @template K The tuple of keys to omit.
     */
    ObjectType.prototype.omit = function () {
        var keys2 = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            keys2[_i] = arguments[_i];
        }
        var keys = this.keys;
        var nextProps = {};
        for (var i = 0; i < keys.length; ++i) {
            if (!keys2.includes(keys[i])) {
                nextProps[keys[i]] = this.valueTypes[i];
            }
        }
        var type = new ObjectType(nextProps, this.indexerType);
        type.keysMode = this.keysMode;
        return type;
    };
    /**
     * Returns an object type that allows only known keys and has no index signature.
     *
     * @returns The modified object type.
     */
    ObjectType.prototype.exact = function (options) {
        var type = cloneObject(this);
        type.keysMode = 0 /* ObjectKeysMode.EXACT */;
        type.exactOptions = options;
        return type;
    };
    /**
     * Returns an object type that doesn't have indexer signature and all unknown keys are stripped.
     *
     * @returns The modified object type.
     */
    ObjectType.prototype.strip = function () {
        var type = cloneObject(this);
        type.keysMode = 1 /* ObjectKeysMode.STRIP */;
        return type;
    };
    /**
     * Returns an object type that has an indexer signature that doesn't constrain values.
     *
     * @returns The modified object type.
     */
    ObjectType.prototype.preserve = function () {
        var type = cloneObject(this);
        type.keysMode = 2 /* ObjectKeysMode.PRESERVE */;
        return type;
    };
    /**
     * Returns an object type that has an indexer signature that is constrained by the given type.
     *
     * @param indexType The type of the indexer values.
     * @returns The modified object type.
     *
     * @template I1 The indexer signature type.
     */
    ObjectType.prototype.index = function (indexType) {
        var type = cloneObject(this);
        type.keysMode = 3 /* ObjectKeysMode.INDEXER */;
        type.indexerType = indexType;
        return type;
    };
    ObjectType.prototype.parse = function (input, options) {
        if (!isObjectLike(input)) {
            raiseIssue(input, 'type', 'object', this.options, 'Must be an object');
        }
        var _a = this, propEntries = _a.propEntries, keysMode = _a.keysMode, keys = _a.keys, indexerType = _a.indexerType;
        var output = input;
        var issues;
        if (keysMode === 0 /* ObjectKeysMode.EXACT */ || keysMode === 1 /* ObjectKeysMode.STRIP */) {
            for (var key in input) {
                if (keys.includes(key)) {
                    continue;
                }
                if (keysMode === 0 /* ObjectKeysMode.EXACT */) {
                    issues = raiseIssuesOrPush(issues, options, input, 'unknownKeys', key, this.exactOptions, 'Must have known keys but found ' + key);
                }
                if (output === input) {
                    output = copyObjectKnownKeys(input, keys);
                }
            }
        }
        if (this.async) {
            var promises = [];
            var objectKeys_1 = keys;
            var handleResults = function (results) {
                for (var i = 0; i < objectKeys_1.length; ++i) {
                    var key = objectKeys_1[i];
                    var outputValue = results[i];
                    if (isEqual(outputValue, input[key])) {
                        continue;
                    }
                    if (output === input) {
                        output = copyObjectEnumerableKeys(input);
                    }
                    output[key] = outputValue;
                }
                return output;
            };
            for (var _i = 0, propEntries_1 = propEntries; _i < propEntries_1.length; _i++) {
                var _b = propEntries_1[_i], key = _b[0], type = _b[1];
                promises.push(parseAsync(type, input[key], options).catch(createCatchForKey(key)));
            }
            if (keysMode === 3 /* ObjectKeysMode.INDEXER */) {
                for (var key in input) {
                    if (keys.includes(key)) {
                        continue;
                    }
                    if (objectKeys_1 === keys) {
                        objectKeys_1 = keys.slice(0);
                    }
                    objectKeys_1.push(key);
                    promises.push(parseAsync(indexerType, input[key], options).catch(createCatchForKey(key)));
                }
            }
            if (isFast(options)) {
                return promiseAll(promises).then(handleResults);
            }
            return promiseAllSettled(promises).then(createValuesExtractor(issues)).then(handleResults);
        }
        for (var _c = 0, propEntries_2 = propEntries; _c < propEntries_2.length; _c++) {
            var _d = propEntries_2[_c], key = _d[0], type = _d[1];
            var value = input[key];
            var outputValue = void 0;
            try {
                outputValue = type.parse(value, options);
            }
            catch (error) {
                issues = raiseIssuesOrCaptureForKey(error, issues, options, key);
            }
            if (isEqual(outputValue, value) || issues !== undefined) {
                continue;
            }
            if (output === input) {
                output = copyObjectEnumerableKeys(input);
            }
            output[key] = outputValue;
        }
        if (keysMode === 3 /* ObjectKeysMode.INDEXER */) {
            for (var key in input) {
                if (keys.includes(key)) {
                    continue;
                }
                var value = input[key];
                var outputValue = void 0;
                try {
                    outputValue = indexerType.parse(value, options);
                }
                catch (error) {
                    issues = raiseIssuesOrCaptureForKey(error, issues, options, key);
                }
                if (isEqual(outputValue, value) || issues !== undefined) {
                    continue;
                }
                if (output === input) {
                    output = copyObjectEnumerableKeys(input);
                }
                output[key] = outputValue;
            }
        }
        raiseIssuesIfDefined(issues);
        return output;
    };
    return ObjectType;
}(Type));
export { ObjectType };
