import { Awaitable, ConstraintOptions, Issue, ParserOptions } from '../shared-types';
/**
 * Infers the type from the type definition.
 *
 * @template X The type definition to infer the type from.
 */
export declare type InferType<X extends AnyType> = X extends Type<infer T> ? T : never;
/**
 * An arbitrary type definition.
 */
export declare type AnyType = Type<any> | Type<never>;
/**
 * The abstract type definition.
 */
export declare abstract class Type<T> {
    readonly async: boolean;
    protected options?: ConstraintOptions | undefined;
    constructor(async: boolean, options?: ConstraintOptions | undefined);
    /**
     * Parses the input so it conforms the type definition.
     *
     * @param input The input to parse.
     * @param options The parser options.
     * @returns The parsed value.
     * @throws {@link ValidationError} if parsing has failed.
     */
    abstract parse(input: unknown, options?: ParserOptions): Awaitable<T>;
    validate(input: unknown, options?: ParserOptions): Issue[] | null;
    validateAsync(input: unknown, options?: ParserOptions): Promise<Issue[] | null>;
    /**
     * Transforms the input to the new type.
     *
     * @param transformer The callback that takes the input and transforms it to the new type.
     * @returns The transformed type.
     *
     * @template O The type of the output input.
     */
    transform<O>(transformer: Transformer<T, O>): TransformedType<this, O>;
    /**
     * Asynchronously transforms the input to the new type.
     *
     * @param transformer The callback that takes the input and transforms it to the new type.
     * @returns The transformed type.
     *
     * @template O The type of the output input.
     */
    transformAsync<O>(transformer: Transformer<T, Promise<O>>): TransformedType<this, O>;
    refine<O extends T>(refiner: (value: T) => value is O, options?: ConstraintOptions): TransformedType<this, O>;
    refine(refiner: (value: T) => unknown, options?: ConstraintOptions): TransformedType<this, T>;
}
/**
 * Transforms the input value to the output value.
 *
 * @param input The input value that must be transformed.
 * @returns The output value.
 *
 * @template I The type of the input value.
 * @template O The type of the output input.
 */
export declare type Transformer<I, O> = (input: I) => O;
/**
 * The transforming type definition.
 *
 * @template X The type definition of the input value.
 * @template T The output value.
 */
export declare class TransformedType<X extends AnyType, O> extends Type<O> {
    protected type: X;
    protected transformer: Transformer<InferType<X>, Awaitable<O>>;
    /**
     * Creates a new {@link TransformedType} instance.
     *
     * @param type The type that parses a transformation input value.
     * @param async `true` if transformer returns a `Promise`, or `false` otherwise.
     * @param transformer The transformer that converts input value to the output value.
     */
    constructor(type: X, async: boolean, transformer: Transformer<InferType<X>, Awaitable<O>>);
    parse(input: unknown, options?: ParserOptions): Awaitable<O>;
}
