import { __extends } from "tslib";
import { extractIssues, parseAsync, raiseIssue, returnNull } from '../utils';
/**
 * The abstract type definition.
 */
var Type = /** @class */ (function () {
    function Type(async, options) {
        this.async = async;
        this.options = options;
    }
    Type.prototype.validate = function (input, options) {
        if (this.async) {
            throw new Error('Cannot use async type');
        }
        try {
            this.parse(input, options);
        }
        catch (error) {
            return extractIssues(error);
        }
        return null;
    };
    Type.prototype.validateAsync = function (input, options) {
        return parseAsync(this, input, options).then(returnNull, extractIssues);
    };
    /**
     * Transforms the input to the new type.
     *
     * @param transformer The callback that takes the input and transforms it to the new type.
     * @returns The transformed type.
     *
     * @template O The type of the output input.
     */
    Type.prototype.transform = function (transformer) {
        return new TransformedType(this, false, transformer);
    };
    /**
     * Asynchronously transforms the input to the new type.
     *
     * @param transformer The callback that takes the input and transforms it to the new type.
     * @returns The transformed type.
     *
     * @template O The type of the output input.
     */
    Type.prototype.transformAsync = function (transformer) {
        return new TransformedType(this, true, transformer);
    };
    Type.prototype.refine = function (refiner, options) {
        return this.transform(function (input) {
            if (!refiner(input)) {
                raiseIssue(input, 'refined', undefined, options, 'Must be refined');
            }
            return input;
        });
    };
    return Type;
}());
export { Type };
/**
 * The transforming type definition.
 *
 * @template X The type definition of the input value.
 * @template T The output value.
 */
var TransformedType = /** @class */ (function (_super) {
    __extends(TransformedType, _super);
    /**
     * Creates a new {@link TransformedType} instance.
     *
     * @param type The type that parses a transformation input value.
     * @param async `true` if transformer returns a `Promise`, or `false` otherwise.
     * @param transformer The transformer that converts input value to the output value.
     */
    function TransformedType(type, async, transformer) {
        var _this = _super.call(this, async || type.async) || this;
        _this.type = type;
        _this.transformer = transformer;
        return _this;
    }
    TransformedType.prototype.parse = function (input, options) {
        var _a = this, type = _a.type, transformer = _a.transformer;
        if (this.async) {
            var promise = type.async ? type.parse(input, options) : parseAsync(type, input, options);
            return promise.then(transformer);
        }
        return transformer(type.parse(input, options));
    };
    return TransformedType;
}(Type));
export { TransformedType };
